# -*- coding: utf-8 -*-
'''
National Instruments SystemLink Assets' Health Monitoring Beacon
'''
# Import Python libs
import asyncio
import logging
import os
import sys
from typing import Any, Dict, List, Tuple, Union, Optional
from urllib.parse import quote

from asset_discovery.models.sensor_model import SensorModel
from asset_discovery.models.sensor_properties import SensorProperties
from asset_discovery.models.asset_health_data import AssetHealthData
from asset_discovery.models.operation_result import OperationResult

# Import salt libs
# pylint: disable=import-error,3rd-party-module-not-gated
import salt.utils

# pylint: enable=import-error,3rd-party-module-not-gated

# Import National Instruments libs
# pylint: disable=import-error,3rd-party-module-not-gated
try:
    # noinspection PyUnresolvedReferences
    from systemlink.messagebus.paths import get_http_master_file
    # noinspection PyUnresolvedReferences
    from systemlink.clientconfig import get_configuration_by_id
    # noinspection PyUnresolvedReferences
    from systemlink.clientconfig.configuration import Configuration
    # noinspection PyUnresolvedReferences
    from systemlink.clients.nitag import (
        ApiClient, ApiException, Tag, TagListAndMergeFlag,
        TagsApi, TagUpdate, TagValue, TimestampedTagValue
    )
    # noinspection PyUnresolvedReferences
    from systemlink.clients.nitag.rest import RESTResponse

    HAS_SYSTEMLINK = True
except ImportError:
    HAS_SYSTEMLINK = False

# Placeholder for the _ni_asset_util_module
ni_asset_util = None  # pylint: disable=invalid-name

__virtualname__ = 'ni_asset_health_monitoring'

# Set up logging
log = logging.getLogger(__name__)

BEACON_INITIALIZED = False
BEACON_INTERVAL_DISCONNECTED = 15
BEACON_INTERVAL_DEFAULT = 300
BEACON_DEFAULT_ENABLED = True

HTTP_RETRY_COUNT = 3
HTTP_RETRY_TIMEOUT_SECONDS = 2

MONITORING_ENABLED_KEY = 'asset_health_monitoring_enabled'
MONITORING_INTERVAL_KEY = 'asset_health_monitoring_interval'

HTTP_MASTER_CONFIGURATION_ID = 'SYSTEMLINK_MASTER'

DEFAULT_ASSET_HEALTH_MONITORING_ENABLED = True
DEFAULT_ASSET_HEALTH_MONITORING_INTERVAL = 300
LAST_WORKSPACE = None
CURRENT_WORKSPACE = None

HEALTH = 'Health'
TEMPERATURE = 'Temperature'
VOLTAGE = 'Voltage'
POWER = 'Power'
READING = 'Reading'
ALARM = 'Alarm'
STATE = 'State'

TYPE_DOUBLE = 'DOUBLE'
TYPE_INT = 'INT'

UNITS_CELSIUS = '°C'
UNITS_VOLTS = 'V'
UNITS_WATTS = 'W'

PROPS_SENSOR_NAME = 'Sensor name'
PROPS_MINION_ID = 'Minion id'
PROPS_SYSTEM_NAME = 'System name'
PROPS_VENDOR_NAME = 'Vendor name'
PROPS_MODEL_NAME = 'Model name'
PROPS_SERIAL_NUMBER = 'Serial number'
PROPS_ASSET_NAME = 'Asset name'

SENSOR_PROPS_SERIALIZATION_KEYS = {
    SensorProperties.Interval: 'Polling interval',
    SensorProperties.LowerUser: 'Lower user threshold',
    SensorProperties.UpperUser: 'Upper user threshold',
    SensorProperties.LowerRisk: 'Lower damage risk threshold',
    SensorProperties.UpperRisk: 'Upper damage risk threshold',
    SensorProperties.LowerCritical: 'Lower critical threshold',
    SensorProperties.UpperCritical: 'Upper critical threshold',
    SensorProperties.Accuracy: 'Accuracy',
    SensorProperties.Nominal: 'Nominal'
}

RESERVED_RETENTION_TYPE_KEY = 'nitagRetention'
RESERVED_RETENTION_DAYS_KEY = 'nitagHistoryTTLDays'
RESERVED_UNITS_KEY = 'units'
RESERVED_HYPERLINK_KEY = 'hyperlink'
RESERVED_DISPLAY_NAME_KEY = 'displayName'

RETENTION_DURATION_DAYS = '365'
RETENTION_TYPE = 'DURATION'

# For firing local events
EVENT = None


class TagWrapper:
    def __init__(self, create_or_update_model: Tag, update_value_model: TagUpdate):
        self._create_or_update_model = create_or_update_model
        self._update_value_model = update_value_model

    @property
    def create_or_update_model(self):
        return self._create_or_update_model

    @property
    def update_value_model(self):
        return self._update_value_model


def _create_and_publish_tags() -> None:
    global HAS_SYSTEMLINK  # pylint: disable=global-statement

    if not HAS_SYSTEMLINK:
        raise Exception("Failed to import SystemLink dependencies.")
    health_data_result = _get_health_data()
    if not health_data_result.status.succeeded:
        raise Exception("Failed to retrieve health information.")
    health_data = health_data_result.content
    tags = _create_tags(health_data)
    if tags is None or len(tags) == 0:
        return
    create_or_update_models = [tag.create_or_update_model for tag in tags]
    tags_and_merge_flag = TagListAndMergeFlag(create_or_update_models, False)
    tag_value_update_models = [tag.update_value_model for tag in tags]
    _publish_tags_with_loop(tags_and_merge_flag, tag_value_update_models)


def _get_health_data() -> OperationResult:
    '''
    Gets health data for assets.
    '''
    return ni_asset_util.get_asset_discovery_engine().get_health_data()


def _publish_tags_with_loop(tags_and_merge_flag: TagListAndMergeFlag, tag_updates: List[TagUpdate]):
    loop = None
    new_event_loop_created = None
    try:
        new_event_loop_created = _ensure_event_loop()
        loop = asyncio.get_event_loop()
        loop.run_until_complete(_publish_tags_over_http(tags_and_merge_flag, tag_updates))
    finally:
        if new_event_loop_created is True and loop is not None:
            loop.stop()
            loop.close()
            asyncio.set_event_loop(None)


def _ensure_event_loop() -> bool:
    try:
        asyncio.get_event_loop()
        return False
    except RuntimeError:
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        return True


async def _publish_tags_over_http(tags_and_merge_flag: TagListAndMergeFlag, tag_updates: List[TagUpdate]):
    global LAST_WORKSPACE
    global CURRENT_WORKSPACE

    writer = TagHttpWriter()
    if writer.get_configuration() is None:
        log.warning("Failed to create the TagHttpWriter with a valid config.")
        return
    if LAST_WORKSPACE != CURRENT_WORKSPACE:
        await writer.create_or_update_tags(tags_and_merge_flag)
        LAST_WORKSPACE = CURRENT_WORKSPACE
    await writer.update_tag_values(tag_updates)


def _create_tags(health_data: List[AssetHealthData]) -> List[TagWrapper]:
    tags = []
    common_properties = _get_common_properties()
    for asset_health_data in health_data:
        tags_for_asset = _create_asset_health_tags(asset_health_data, common_properties)
        tags.extend(tags_for_asset)
    return tags


def _get_common_properties() -> Dict[str, str]:
    return {
        PROPS_SYSTEM_NAME: __grains__['host'],
        PROPS_MINION_ID: __grains__['id'],
        RESERVED_RETENTION_TYPE_KEY: RETENTION_TYPE,
        RESERVED_RETENTION_DAYS_KEY: RETENTION_DURATION_DAYS
    }


def _create_asset_health_tags(
        asset_health_data: AssetHealthData,
        common_properties: Dict[str, str]) -> List[TagWrapper]:
    asset_properties = _get_asset_properties(asset_health_data)
    base_properties = {**common_properties, **asset_properties}
    path_prefix = _get_asset_health_path_prefix(asset_health_data)
    return _create_asset_sensor_tags(asset_health_data, base_properties, path_prefix)


def _get_asset_properties(asset_health_data: AssetHealthData) -> Dict[str, str]:
    props = {}
    identification = asset_health_data.asset_identification

    vendor = identification.vendor_name
    if vendor is not None:
        props[PROPS_VENDOR_NAME] = vendor

    model = identification.model_name
    if model is not None:
        props[PROPS_MODEL_NAME] = model

    serial_number = identification.serial_number
    props[PROPS_SERIAL_NUMBER] = serial_number

    props[PROPS_ASSET_NAME] = asset_health_data.asset_name
    return props


def _get_asset_health_path_prefix(asset_health_data: AssetHealthData) -> str:
    identification = asset_health_data.asset_identification
    vendor = identification.vendor_name
    if vendor is None:
        vendor = str(identification.vendor_number)
    model = identification.model_name
    if model is None:
        model = str(identification.model_number)
    serial_number = str(identification.serial_number)
    vendor = _sanitize_path_component(vendor)
    model = _sanitize_path_component(model)
    serial_number = _sanitize_path_component(serial_number)
    return "{0}.{1}.{2}.{3}".format(vendor, model, serial_number, HEALTH)


def _create_asset_sensor_tags(
        asset_health_data: AssetHealthData,
        base_properties: Dict[str, str],
        path_prefix: str) -> List[TagWrapper]:
    tags = []
    temperature_tags = _create_temp_tags(asset_health_data.temperature_sensors, base_properties, path_prefix)
    voltage_tags = _create_volt_tags(asset_health_data.voltage_sensors, base_properties, path_prefix)
    power_tags = _create_power_tags(asset_health_data.power_sensors, base_properties, path_prefix)
    tags.extend(temperature_tags)
    tags.extend(voltage_tags)
    tags.extend(power_tags)
    return tags


def _create_temp_tags(
        temp_sensors: List[SensorModel],
        base_properties: Dict[str, str],
        path_prefix: str) -> List[TagWrapper]:
    path_prefix = "{0}.{1}".format(path_prefix, TEMPERATURE)
    if temp_sensors is None:
        return []
    tags = []
    for sensor in temp_sensors:
        tags.extend(_create_temp_sensor_tags(sensor, base_properties, path_prefix))
    return tags


def _create_temp_sensor_tags(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> List[TagWrapper]:
    reading_tag = _create_temp_reading_tag(sensor, base_properties, path_prefix)
    alarm_tag = _create_temp_alarm_tag(sensor, base_properties, path_prefix)
    state_tag = _create_temp_state_tag(sensor, base_properties, path_prefix)
    sensor_tags = [reading_tag, alarm_tag, state_tag]
    return [t for t in sensor_tags if t is not None]


def _create_temp_reading_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> Optional[TagWrapper]:
    optional_sensor_props_keys = [
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
        SensorProperties.Accuracy
    ]
    value = sensor.reading
    path_prefix = "{0}.{1}".format(path_prefix, READING)
    units = UNITS_CELSIUS
    tag_type = TYPE_DOUBLE
    display_name = _get_sensor_reading_display_name(sensor, base_properties)
    return _create_sensor_tag(
        sensor, value, tag_type, display_name, base_properties, path_prefix, optional_sensor_props_keys, units)


def _get_sensor_reading_display_name(sensor: SensorModel, base_properties: Dict[str, str]) -> str:
    asset_name = base_properties[PROPS_ASSET_NAME]
    serial_number = base_properties[PROPS_SERIAL_NUMBER]
    return "{0} ({1}) - {2}".format(asset_name, serial_number, sensor.name)


def _create_sensor_tag(
        sensor: SensorModel,
        value: Optional[Any],
        tag_type: str,
        display_name: str,
        base_properties: Dict[str, Any],
        path_prefix: str,
        optional_props_keys: List[str],
        units: Optional[str]) -> Optional[TagWrapper]:
    global CURRENT_WORKSPACE
    if value is None:
        return None
    value = str(value)
    sensor_name = _sanitize_path_component(sensor.name)
    path = "{0}.{1}".format(path_prefix, sensor_name)
    hyperlink = _get_hyperlink(path, CURRENT_WORKSPACE)
    sensor_properties = _get_sensor_properties(sensor, optional_props_keys)
    props = {**base_properties, **sensor_properties,
             RESERVED_DISPLAY_NAME_KEY: display_name, RESERVED_HYPERLINK_KEY: hyperlink}
    if units is not None:
        props[RESERVED_UNITS_KEY] = units
    return _create_tag_wrapper(path, value, props, tag_type)


def _get_hyperlink(path: str, workspace: str) -> str:
    '''
    Return a relative URL to the tag details page for the given tag.

    :param path: The tag path to generate a relative URL for.
    :param workspace: The workspace of the system.
    :return: A relative URL to the tag details page for the given tag.
    '''
    return '#tagviewer/tag/' + quote(workspace, safe='') + '/' + quote(path, safe='')


def _get_sensor_properties(sensor: SensorModel, optional_props_keys: List[str]) -> Dict[str, Any]:
    props = {
        PROPS_SENSOR_NAME: sensor.name
    }
    for key in optional_props_keys:
        if key in sensor.properties:
            serialization_key = SENSOR_PROPS_SERIALIZATION_KEYS[key]
            value = sensor.properties[key]
            props[serialization_key] = str(value)
    return props


def _create_tag_wrapper(path: str, value: str, properties: Dict[str, Any], tag_type: str) -> TagWrapper:
    tag_create_or_update_model = Tag(type=tag_type, properties=properties, path=path, collect_aggregates=True)
    # A timestamp of ``None`` means use the server time.
    tag_updated_value = TimestampedTagValue(value=TagValue(value=value, type=tag_type), timestamp=None)
    tag_update_model = TagUpdate(path=path, updates=[tag_updated_value])
    return TagWrapper(tag_create_or_update_model, tag_update_model)


def _create_temp_alarm_tag(
        sensor: SensorModel,
        base_properties: Dict[str, Any],
        path_prefix: str) -> Optional[TagWrapper]:
    optional_sensor_props_keys = [
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
    ]
    value = sensor.alarm
    path_prefix = "{0}.{1}".format(path_prefix, ALARM)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_alarm_display_name(sensor, base_properties)
    return _create_sensor_tag(
        sensor, value, tag_type, display_name, base_properties, path_prefix, optional_sensor_props_keys, units)


def _get_sensor_alarm_display_name(sensor: SensorModel, base_properties: Dict[str, str]) -> str:
    display_name = _get_sensor_reading_display_name(sensor, base_properties)
    return "{0} alarm".format(display_name)


def _create_temp_state_tag(
        sensor: SensorModel,
        base_properties: Dict[str, any],
        path_prefix: str) -> Optional[TagWrapper]:
    optional_sensor_props_keys = [
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
        SensorProperties.Interval
    ]
    value = sensor.state
    path_prefix = "{0}.{1}".format(path_prefix, STATE)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_state_display_name(sensor, base_properties)
    return _create_sensor_tag(
        sensor, value, tag_type, display_name, base_properties, path_prefix, optional_sensor_props_keys, units)


def _get_sensor_state_display_name(sensor: SensorModel, base_properties: Dict[str, str]) -> str:
    display_name = _get_sensor_reading_display_name(sensor, base_properties)
    return "{0} state".format(display_name)


def _create_volt_tags(
        volt_sensors: List[SensorModel],
        base_properties: Dict[str, str],
        path_prefix: str) -> List[TagWrapper]:
    path_prefix = "{0}.{1}".format(path_prefix, VOLTAGE)
    if volt_sensors is None:
        return []
    tags = []
    for sensor in volt_sensors:
        tags.extend(_create_volt_sensor_tags(sensor, base_properties, path_prefix))
    return tags


def _create_volt_sensor_tags(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> List[TagWrapper]:
    reading_tag = _create_volt_reading_tag(sensor, base_properties, path_prefix)
    alarm_tag = _create_volt_alarm_tag(sensor, base_properties, path_prefix)
    state_tag = _create_volt_state_tag(sensor, base_properties, path_prefix)
    sensor_tags = [reading_tag, alarm_tag, state_tag]
    return [t for t in sensor_tags if t is not None]


def _create_volt_reading_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> Optional[TagWrapper]:
    optional_sensor_props_keys = [
        SensorProperties.Nominal,
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
        SensorProperties.Accuracy
    ]
    value = sensor.reading
    path_prefix = "{0}.{1}".format(path_prefix, READING)
    units = UNITS_VOLTS
    tag_type = TYPE_DOUBLE
    display_name = _get_sensor_reading_display_name(sensor, base_properties)
    return _create_sensor_tag(
        sensor, value, tag_type, display_name, base_properties, path_prefix, optional_sensor_props_keys, units)


def _create_volt_alarm_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> Optional[TagWrapper]:
    optional_sensor_props_keys = [
        SensorProperties.Nominal,
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser
    ]
    value = sensor.alarm
    path_prefix = "{0}.{1}".format(path_prefix, ALARM)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_alarm_display_name(sensor, base_properties)
    return _create_sensor_tag(
        sensor, value, tag_type, display_name, base_properties, path_prefix, optional_sensor_props_keys, units)


def _create_volt_state_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> Optional[TagWrapper]:
    optional_sensor_props_keys = [
        SensorProperties.Nominal,
        SensorProperties.UpperRisk,
        SensorProperties.LowerRisk,
        SensorProperties.UpperCritical,
        SensorProperties.LowerCritical,
        SensorProperties.UpperUser,
        SensorProperties.LowerUser,
        SensorProperties.Interval
    ]
    value = sensor.state
    path_prefix = "{0}.{1}".format(path_prefix, STATE)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_state_display_name(sensor, base_properties)
    return _create_sensor_tag(
        sensor, value, tag_type, display_name, base_properties, path_prefix, optional_sensor_props_keys, units)


def _create_power_tags(
        power_sensors: List[SensorModel],
        base_properties: Dict[str, str],
        path_prefix: str) -> List[TagWrapper]:
    path_prefix = "{0}.{1}".format(path_prefix, POWER)
    if power_sensors is None:
        return []
    tags = []
    for sensor in power_sensors:
        tags.extend(_create_power_sensor_tags(sensor, base_properties, path_prefix))
    return tags


def _create_power_sensor_tags(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> List[TagWrapper]:
    reading_tag = _create_power_reading_tag(sensor, base_properties, path_prefix)
    alarm_tag = _create_power_alarm_tag(sensor, base_properties, path_prefix)
    state_tag = _create_power_state_tag(sensor, base_properties, path_prefix)
    sensor_tags = [reading_tag, alarm_tag, state_tag]
    return [t for t in sensor_tags if t is not None]


def _create_power_reading_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> Optional[TagWrapper]:
    optional_sensor_props_keys = [
        SensorProperties.UpperCritical,
        SensorProperties.UpperRisk,
        SensorProperties.UpperUser,
        SensorProperties.Accuracy
    ]

    value = sensor.reading
    path_prefix = "{0}.{1}".format(path_prefix, READING)
    units = UNITS_WATTS
    tag_type = TYPE_DOUBLE
    display_name = _get_sensor_reading_display_name(sensor, base_properties)
    return _create_sensor_tag(sensor, value, tag_type, display_name, base_properties, path_prefix,
                              optional_sensor_props_keys, units)


def _create_power_alarm_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> Optional[TagWrapper]:
    optional_sensor_props_keys = [
        SensorProperties.UpperCritical,
        SensorProperties.UpperRisk,
        SensorProperties.UpperUser,
    ]
    value = sensor.alarm
    path_prefix = "{0}.{1}".format(path_prefix, ALARM)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_alarm_display_name(sensor, base_properties)
    return _create_sensor_tag(sensor, value, tag_type, display_name, base_properties, path_prefix,
                              optional_sensor_props_keys, units)


def _create_power_state_tag(
        sensor: SensorModel,
        base_properties: Dict[str, str],
        path_prefix: str) -> Optional[TagWrapper]:
    optional_sensor_props_keys = [
        SensorProperties.UpperCritical,
        SensorProperties.UpperRisk,
        SensorProperties.UpperUser,
        SensorProperties.Interval
    ]
    value = sensor.state
    path_prefix = "{0}.{1}".format(path_prefix, STATE)
    units = None
    tag_type = TYPE_INT
    display_name = _get_sensor_state_display_name(sensor, base_properties)
    return _create_sensor_tag(sensor, value, tag_type, display_name, base_properties, path_prefix,
                              optional_sensor_props_keys, units)


def _sanitize_path_component(component: str) -> str:
    return component.replace('.', '\u2024')


def __virtual__() -> Union[str, Tuple[bool, str]]:
    '''
    During module lazy loading, will return the virtual name of this module.

    :return: The virtual name of this module. On error, return a 2-tuple with
        ``False`` for the first item and the error message for the second.
    '''
    _import_modules()
    return __virtualname__


def _import_modules():
    # Add extra search paths to load 'ni_asset_util'
    paths_to_add = __opts__['module_dirs']
    extmods_module_dir = os.path.join(__opts__['extension_modules'], 'modules')
    if extmods_module_dir not in paths_to_add:
        paths_to_add.append(extmods_module_dir)
    sys.path.extend(paths_to_add)
    global ni_asset_util  # pylint: disable=global-statement
    try:
        import _ni_asset_util as ni_asset_util
    finally:
        # Remove the extra search paths that we added to sys.path
        for path in paths_to_add:
            sys.path.remove(path)


def validate(config: List[Dict[str, Any]]) -> Tuple[bool, str]:
    '''
    Validate the beacon configuration

    :param config: The beacon configuration.
    :return: A 2-tuple. The first item is ``True`` if the
        configuration is valid and ``False`` otherwise. The second item
        is a message.
    '''
    if not isinstance(config, list):
        return False, ('Configuration for ni_asset_health_monitoring '
                       'beacon must be a list.')
    return True, 'Valid beacon configuration'


def beacon(config: List[Dict[str, Any]]) -> List[Dict[str, Any]]:  # pylint: disable=unused-argument
    '''
    National Instruments SystemLink Assets' Health Monitoring Beacon

    :param config: The beacon configuration.
    :return: A list of events to send to the master.
    '''
    global BEACON_INITIALIZED  # pylint: disable=global-statement
    global CURRENT_WORKSPACE  # pylint: disable=global-statement

    try:
        if not BEACON_INITIALIZED:
            success = _init_beacon()
            if not success:
                return []
        CURRENT_WORKSPACE = __grains__.get('systemlink_workspace', None)
        if CURRENT_WORKSPACE is None:
            _disconnected_beacon_interval()
        else:
            _connected_beacon_interval()
            _create_and_publish_tags()
    except Exception as exc:  # pylint: disable=broad-except
        log.error(
            'Unexpected exception in "ni_asset_health_monitoring beacon": %s',
            exc, exc_info=True
        )
        _disconnected_beacon_interval()

    return []


def _init_beacon() -> bool:
    '''
    Initialize the beacon

    :return: ``True`` if initialized successfully, ``False`` otherwise.
    '''
    global BEACON_INITIALIZED  # pylint: disable=global-statement

    enabled = __grains__.get(MONITORING_ENABLED_KEY, BEACON_DEFAULT_ENABLED)
    if __opts__.get('master_type') == 'disable' or not enabled:
        _fire_local_event(
            {
                'func': 'disable_beacon',
                'name': __virtualname__
            },
            'manage_beacons'
        )
        return False

    BEACON_INITIALIZED = True
    return True


def _fire_local_event(event_data: Dict[str, Any], event_tag: str) -> None:
    '''
    Fire an event to the local event bus.

    :param event_data: The data associated with the event.
    :param event_tag: The identifier of the event.
    '''
    global EVENT  # pylint: disable=global-statement

    if not EVENT:
        EVENT = salt.utils.event.get_event(
            'minion', opts=__opts__, listen=False
        )
    EVENT.fire_event(event_data, event_tag)


def _connected_beacon_interval() -> None:
    '''
    This is invoked once we successfully connected
    and want to ensure that the long term beacon
    interval is in effect.
    '''
    global BEACON_INTERVAL_DEFAULT

    interval = __grains__.get(MONITORING_INTERVAL_KEY, BEACON_INTERVAL_DEFAULT)
    current_interval = __opts__['beacons'][__virtualname__][0].get('interval', BEACON_INTERVAL_DEFAULT)
    if current_interval != interval:
        _set_beacon_interval(interval)


def _disconnected_beacon_interval() -> None:
    '''
    This is invoked when we have not successfully connected
    or have been disconnected due to error and want to ensure
    that the short term beacon interval is in effect.
    '''
    global BEACON_INITIALIZED
    global BEACON_INTERVAL_DEFAULT

    current_interval = __opts__['beacons'][__virtualname__][0].get('interval', BEACON_INTERVAL_DEFAULT)
    if current_interval != BEACON_INTERVAL_DISCONNECTED:
        _set_beacon_interval(BEACON_INTERVAL_DISCONNECTED)


def _set_beacon_interval(interval: int) -> None:
    '''
    This is invoked once we successfully connected
    and want to ensure that the long term beacon
    interval is in effect.

    :param interval: The interval on which to run the beacon.
    '''
    _fire_local_event(
        {
            'func': 'modify',
            'name': __virtualname__,
            'beacon_data': [{
                'interval': interval
            }]
        },
        'manage_beacons'
    )


class TagHttpWriter:  # pylint: disable=too-few-public-methods
    '''
    Abstraction over HTTP connection which facilitates communication
    between the minion and the Tag service.
    '''

    def __init__(self) -> None:
        '''
        Initialize the message service if not already initialized
        '''
        self._configuration = None
        self._setup_communication()

    def get_configuration(self) -> Configuration:
        return self._configuration

    def _setup_communication(self) -> None:
        '''
        Set up the communication.
        '''
        file_path = get_http_master_file()
        if not os.path.isfile(file_path):
            # The Skyline Master file is not available.
            # Can't set up the Message Service without it.
            return

        self._setup_http_connection()

    def _setup_http_connection(self) -> None:
        '''
        Set up the HTTP communication and initialize the Tag API object.
        '''
        tag_route_name = '/nitag'
        self._configuration = get_configuration_by_id(
            HTTP_MASTER_CONFIGURATION_ID,
            tag_route_name,
            False)

    async def create_or_update_tags(
            self,
            tags_and_merge: TagListAndMergeFlag):
        '''
        Publish an /update-tags request over HTTP.

        :param tags_and_merge: An instance of TagListAndMergeFlag representing the tags and the merge properties flag.
        '''
        async with ApiClient(self._configuration) as api_client:
            tags_api = TagsApi(api_client=api_client)
            response = await tags_api.create_or_update_tags(tags_and_merge, _preload_content=False)
            if response.status not in (200, 201, 202):
                data = await response.text()
                rest_response = RESTResponse(response, data)
                raise ApiException(http_resp=rest_response)

    async def update_tag_values(self, updates: List[TagUpdate]) -> None:
        '''
        Publish a /update-current-value request over HTTP.

        :param updates: A list of TagUpdate objects representing the new tag values.
        '''
        async with ApiClient(self._configuration) as api_client:
            tags_api = TagsApi(api_client=api_client)
            response = await tags_api.update_tag_current_values(updates, _preload_content=False)
        if response.status not in (200, 202):
            data = await response.text()
            rest_response = RESTResponse(response, data)
            raise ApiException(http_resp=rest_response)
